#! /usr/bin/env python

from utils import *
import numpy as np
from copy import copy
from test import *
from planner import Planner, Trajectory
import time
import DecMCTS

# Assuming that time between contours is 3 seconds
transfer_time = 2.0


class State:
    """
    State of a robot arm at any node in the MCTS tree
    It stores actions taken so far + any other info which may
        be useful to calculate reward and available actions
    """

    def __init__(self, trajs, total_time, idle_time):
        self.trajs = trajs          # List of Trajectories executed 
        self.time = total_time      # List of cumulative time at the end of each traj
        self.idle_time = idle_time  # Total time spend not printing so far


def state_storer(data, parent_state, action):
    """
    Create State Objects to store in new nodes
    """

    # if root node
    if parent_state==None:
        return State([], [], 0)

    state = State(copy(parent_state.trajs), copy(parent_state.time), parent_state.idle_time)
    state.trajs.append(action)
    state.time.append(action.time[-1] +
                      (0 if len(state.time)==0  else transfer_time + state.time[-1]))

    # if action was a wait, add it to the idle_time
    state.idle_time += (action.time[-1] if action.contour==None else 0)

    # Only add transfer time if it isn't the first action of the robot
    state.idle_time += (transfer_time if len(state.time)!=1 else 0)
    return state


# Reward function needed for DecMCTS
def reward(data, states):
    """
    The reward function is inversely proportional to the idle_time
    This includes the expected idle_time if one of the robot were to
    complete
    """

    # Total waiting times so far
    idle_time = 0.01  #in case idle time ==0 WARNING: this gives a massive reward 
                      # should only occur if total contours= number of robots
    idle_time += sum([states[robot].idle_time for robot in states])

    # Get time contours which still havent been planned
    planned_contours = [traj.contour 
                     for robot in states
                     for traj in states[robot].trajs]
    time_for_unplanned = [c.time[-1]
                              for c in data["contours"]
                              if c not in planned_contours]

    # add wait times from unplanned path
    idle_time += transfer_time*(max(0,len(time_for_unplanned)-1))

    time_for_unplanned = sum(time_for_unplanned) + transfer_time*(max(0,len(time_for_unplanned)-1))

    # Wait times till print finished by slowest arm
    robot_times = [(states[robot].time[-1] if len(states[robot].time)!=0 else 0) for robot in states]
    total_time = max(robot_times) + time_for_unplanned + (0 if max(robot_times)==0 or time_for_unplanned==0 else transfer_time)
    idle_time += sum(total_time - np.array(robot_times)) - time_for_unplanned
    return data["worst_time"]/idle_time


# available actions function needed for DecMCTS
def available_actions(data, states, robot_id):

    # Get all options - Only works cos robot ids in DecMCTS match Planner r1/r2 objects
    options = copy(data["traj"][robot_id]) #remember not to edit the data

    # Remove contours that have already been done
    done_contours = [traj.contour 
                     for robot in states
                     for traj in states[robot].trajs]
    options[:] = [o for o in options if o.contour not in done_contours]

    # if all contours are done
    if len(options)==0:
        return []
    ## Remove contours which would be colliding with existing plans
    # Start time is when last trajectory finished + transfer time
    start_time = (0 if len(states[robot_id].time)==0 else states[robot_id].time[-1] + transfer_time)
    # End time is start_time + maximum of trajectory option lengths
    end_time = start_time + max([o.time[-1] for o in options])

    # Get the trajectories that over lap, and the difference in start times
                #i = np.searchsorted(states[robot].time, start_time)

    overlapping_trajs = [(traj, (states[robot].time[i] - traj.time[-1]) - start_time)
                         for robot in states
                         for i, traj in enumerate(states[robot].trajs)
                         if robot!=robot_id and 
                            (start_time <= states[robot].time[i] and
                            end_time >= states[robot].time[i] - traj.time[-1])]

    options_to_remove = []
    for o in options:
        #for each overlapping trajectory
        for t in overlapping_trajs:

            #If there's a collision at the given time difference
            if planner.CollisionCheckTrajs(o, t[0], t[1]):
                options_to_remove.append(o)
                break

    # this needs to be done separately because cannot remove while iterating same list
    [options.remove(o) for o in options_to_remove]

    # Add a wait 5 seconds option (1 Trajectory + 2x2 transfer time)
    options.append(Trajectory([[0,0,0,0,0,0]], [1], robot_id))

    return options


if __name__=="__main__":

    planner = Planner("env.xml")
    #planner.load_contours("benchy.gcode")
    planner.load_contours([contourIK1, contourIK2])
    planner.calc_trajectories() #get trajectories for every contour

    planner.viz()

    show_contour(contourIK1, planner.env)
    show_contour(contourIK2, planner.env)

    #start = time.time();planner.CollisionCheckTrajs(trajs1[0], trajs2[0], 0);end=time.time()
    # For 50 overlapping waypoints, it takes about 0.03 seconds to compare

    ## DecMCTS stuff
    comm_n=3
    time_for_single_nozzle_print = sum([c.time[-1] for c in planner.contours]) + transfer_time * (len(planner.contours)-1)
    data = {
        "traj": planner.trajs,
        "contours": planner.contours,
        "worst_time": time_for_single_nozzle_print
    }

    # Testing state_storer
    ## R1
    root_state1 = state_storer(data, None, None)
    next_state1 = state_storer(data, root_state1, planner.trajs[planner.r1][-1])
    last_state1 = state_storer(data, next_state1, planner.trajs[planner.r1][0])

    ## R2
    root_state2 = state_storer(data, None, None)
    next_state2 = state_storer(data, root_state2, planner.trajs[planner.r2][0])
    last_state2 = state_storer(data, next_state2, planner.trajs[planner.r2][-1])

    # Testing reward
    states1 = {planner.r1:next_state1, planner.r2:next_state2}
    states2 = {planner.r1:last_state1, planner.r2:root_state2}
    states3 = {planner.r1:root_state1, planner.r2:root_state2}
    states4 = {planner.r1:root_state1, planner.r2:next_state2}
    rew1 = reward(data, states1)
    rew2 = reward(data, states2)
    rew3 = reward(data, states3)
    rew4 = reward(data, states4)

    # Testing Available actions
    ava1 = available_actions(data, states1, planner.r1)
    ava2 = available_actions(data, states2, planner.r1)
    ava3 = available_actions(data, states3, planner.r2)
    ava4 = available_actions(data, states4, planner.r2)

    #create DecMCTS trees for each robot
    tree1 = DecMCTS.Tree(data, reward, available_actions, state_storer, comm_n, planner.r1)
    tree2 = DecMCTS.Tree(data, reward, available_actions, state_storer, comm_n, planner.r2)

    for i in range(10):
        tree1.grow()
        tree2.grow()
        tree1.receive_comms(tree2.send_comms(), tree2.id)
        tree2.receive_comms(tree1.send_comms(), tree1.id)
