from openravepy import *
import numpy as np
from utils import *
import gcode2contour as gc


class Trajectory:
    """
    Stores relevant info about a trajectory:
    - Robot doing them
    - List of poses
    - Time at each waypoint
    - Contour it completes (if it is for a contour)

    funtionality:
    - Animate/Execute using show
    """

    def __init__(self, pose, time, robot, contour=None):
        self.pose = pose
        self.time = time
        self.robot = robot
        self.contour = contour

    def show(self):
        i=1
        while i<len(self.time):
            self.robot.SetDOFValues(self.pose[i-1])
            sleep(self.time[i] - self.time[i-1])
            i +=1

    def __getitem__(self, i):
        return {"time": self.time[i], "pose":self.pose[i]}

    def __len__(self):
        return len(self.time)


class Planner:
    """
    Currently this planner only supports 2 arms, but is being
        built with more in mind.

    To initiate, provide env.xml file
    - It will generate tables and move the robots
    """

    def __init__(self, envfile):
        # Load Environment
        self.env = Environment()
        self.env.Load(envfile)

        # Change collision checker to ode
        self.env.SetCollisionChecker(RaveCreateCollisionChecker(self.env, 'ode'))

        # Create table and place the robots
        body = RaveCreateKinBody(self.env,'')
        body.SetName('table')
        body.InitFromBoxes(numpy.array([[0,0,-0.1,1.5,1.5,0.08]]), True) 
            #0.08 to prevent base collision with table
        self.env.AddKinBody(body)

        # get robots for easy access
        self.r1 = self.env.GetRobots()[0]
        self.r2 = self.env.GetRobots()[1]

        # place robots
        self.r1.SetTransform([1,0,0,0,0,0.75,0])
        self.r2.SetTransform([1,0,0,0,0,-0.75,0])

        self.home_pose = [-pi/2, -pi/2, 0, -pi/2, 0, 0] #can change
        self.solver = {} # Solvers
        self.Tr = {} # Robot pose transforms
        
        for r in self.env.GetRobots():
            self.home(r)
            self.solver[r] = ur_kinematics("UR10")
            self.Tr[r] = matrixFromPose(r.GetTransformPose())

        # trajectories not calculated yet
        self.trajs = None

    def home(self, robot):
        robot.SetDOFValues(self.home_pose)


    def load_contours(self, data):
        """
        If a string is provided, 'data' is file address
         - contours loaded from gcode file
        Else, data must be list of contours, saved directly
        """
        if isinstance(data, basestring):
            self.contours = gc.decode_gcode(gcode_file)
        else:
            self.contours = data


    def viz(self):
        self.env.SetViewer('qtcoin')


    def contour2traj(self, contour, robot, mm=False):
        """
        contour2traj
        - Convert contour and robot to a list of possible trajectories to take

        Inputs
        - contour object (gcode2contour library)
        - robot object (openrave)

        Output:
        - list of contour following Trajectory objects
        """

        # convert mm to m
        if mm:
            pos = [(p[0]/1000, p[1]/1000, p[2]/1000) for p in contour.pos]
        else:
            pos = contour.pos


        trajs = []
        # for each waypoint in contour
        for i in range(len(pos)):

            # Get 8 solutions
            sols = self.solver[robot].ik(world2printTransform(list(pos[i]), self.Tr[robot]))

            # If impossible IK, this robot cannot do this contour
            if len(sols)==0:
                trajs = []
                break

            # Remove invalid solutions (collision with table)
            sols = [s for s in sols if robot.SetDOFValues(s)==None and not self.env.CheckCollision(robot)]
            if i==0:
                # create trajectories with each valid solution
                trajs = [Trajectory([s], [contour.time[0]], robot, contour) for s in sols]

                continue

            # For each trajectory
            for traj in trajs:

                # find closest solution for next waypoint
                sol_dists = [max(np.array(traj.pose[-1])-np.array(s)) for s in sols]

                # if none is close enough, delete this trajectory
                # TODO CANNOT REMOVE TODO TRAJ WHILE TODO ITERATING THROUGH TODO IT
                if min(sol_dists)>0.1:
                    trajs.remove(traj)
                    continue
                else:
                    traj.pose.append(sols[np.argmin(sol_dists)])
                    traj.time.append(contour.time[i])

                # TO DO if two are close, clone trajectory to consider both pathways??
        
        # reset robot to home position so it doesn't intefere with future planning
        self.home(robot)

        return trajs


    def calc_trajectories(self):
        """
        Get the trajectories FOR ALL CONTOURS FOR ALL ROBOTS
        """
        # trajectories for all contours for a give robot will be flattened into single list
        self.trajs = {}
        # Get Trajectories for all robots
        for r in self.env.GetRobots():
            self.trajs[r] = [] 
            for c in self.contours:
                self.trajs[r] = self.trajs[r] + self.contour2traj(c,r)


    def get_trajectories(self):
        """
        This function ensures that the trajectories have been calculated
        when they are called.
        If they need to re re-calculated, call calc_trajectories func directly
        """
        if self.trajs ==None:
            self.calc_trajectories()
        return self.trajs


    def CollisionCheckTrajs(self, traj1, traj2, time_delta):
        """
        Given two trajectories and the difference between their 
            start times, calculate whether or not they are in collision.
        
        NOTE: This function does not ensure that their are enough 
            waypoints along the trajectories for the checkiing to be 
            a sure thing. Make sure there is at least a waypoint every
            5cm minimum.

        Inputs:
        - traj1: Trajectory object
        - traj2: Trajectory object
        - time_delta: float - how much time after traj1 does traj2 start
            (negative if traj2 starts earlier)

        Outputs:
        - Bool - True if Collision happens
        """

        # Add the time delta
        t2time = [t+time_delta for t in traj2.time]

        # for each waypoint in traj1
        for i in range(len(traj1.time)):

            # find the two corresponding waypoints in traj2
            j = np.searchsorted(t2time, traj1.time[i])

            # If time not in overlappnig range, skip
            if j==0 or j==len(t2time):
                continue
            
            # Collision check between traj1 waypoint and interpolated position of traj2
            traj1.robot.SetDOFValues(traj1.pose[i])
            traj2.robot.SetDOFValues(interp_pose(traj2.pose[j-1], traj2.pose[j], t2time[j-1], t2time[j], traj1.time[i]))
            if self.env.CheckCollision(traj1.robot):
                self.home(traj1.robot)
                self.home(traj2.robot)
                return True

        # If no collisions found, Return False
        self.home(traj1.robot)
        self.home(traj2.robot)
        return False
